#include <mpi.h>
#include <vector>
#include "config.h"
#include <algorithm>
#include "DataFlow.hpp"
#include "DetectCollectives.hpp"

// some trickery to get uint64_t --- you would think the natural
// way to do this would be to check for csdtint in the configure
// script and then just include it, however, cstdint only works
// with -std=c++0x at the moment --- this would crush all hopes
// for portable code

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif


#define NDEBUG

bool TupleSourceRankSorter (const tuple* a, const tuple* b) {
	if(a->sowner == b->sowner)
		return a->owner < b->owner;
	else
		return a->sowner < b->sowner;
}


static void RemoveAllTuplesInAfromB(std::vector<tuple*>* A, std::vector<tuple*>* B) {

	// cant use sort because of the null tuple
	//std::sort(A->begin(), A->end(), TupleSourceRankSorter);
	//std::sort(B->begin(), B->end(), TupleSourceRankSorter);

	for (std::vector<tuple*>::iterator a=A->begin(); a!=A->end(); ++a) {
		for (std::vector<tuple*>::iterator b=B->begin(); b!=B->end(); ++b) {
			if (*a == *b) {
				B->erase(b);
				break;
			}
		}
	}

}


std::vector<tuple*>* DetectCollectives(int num_procs, std::vector<tuple*>* tuples, std::vector<Collective*>* colls) {

	// sort
	std::sort(tuples->begin(), tuples->end(), TupleSourceRankSorter);

	{ /******* collective detection - ALLTOALL (and ALLGATHER) - needs relation $(i, o_{i,j}, j, o_{i,j}, s) 0 <= i,j < P, j\neq i$  ******/

		// we need to determine all possible sizes to test for first -- for
		// alltoall, simply use the src=0, dst=0 tuple!
		std::vector<int> candidate_sizes;
		for(std::vector<tuple*>::iterator ri = tuples->begin(); ri!=tuples->end(); ++ri) {
			tuple *t = *ri;
			if(0 == t->owner && 0 == t->sowner) candidate_sizes.push_back(t->size);
		}
		std::sort(candidate_sizes.begin(), candidate_sizes.end());
		candidate_sizes.erase(std::unique(candidate_sizes.begin(), candidate_sizes.end()), candidate_sizes.end());

		for(std::vector<int>::iterator si = candidate_sizes.begin(); si!=candidate_sizes.end(); ++si) {
			// allocating P^2 tuple vector to store found tuples -- NOT scalable!
			std::vector< std::vector<tuple*> > alltoall;
			alltoall.resize(num_procs);
			int size=0, i=0, j=0;


#ifndef NDEBUG
			printf("searching size %i\n", *si);
#endif
			int found = 0;
			for(std::vector<tuple*>::iterator ri = tuples->begin(); ri!=tuples->end(); ++ri) {
				tuple *t = *ri;
				// if we found a suitable alltoall-tuple, add it to the list and
				// increase j (or i)
				//printf("i %i sowner %i j %i owner %i\n", i, t->sowner, j, t->owner);
				if(i == t->sowner && j == t->owner && t->size == *si && t->size == t->ssize) {
					alltoall[i].push_back(t);
					j++; found++;
					if(j == num_procs) { j=0; i++; }
				}
			}
#ifndef NDEBUG
			printf("p: %i, size: %i, found %i\n", num_procs, *si, found);
#endif
			if(found == num_procs*num_procs) {
				// we found an alltoall -- this could, however, also be an
				// allgather if all contributed buffers from each source are
				// identical!
				int allgather=1;
				uint64_t buf;
				for(int i=0; i<num_procs; ++i) {
					buf = alltoall[i][0]->sstart; // reset for next set of hosts
					for(int j=0; j<num_procs; ++j) {
						if(alltoall[i][j]->sstart != buf) allgather = 0;
					}
				}
				if(allgather == 1) ;
#ifndef NDEBUG
					printf("found allgather of size %i!\n", *si);
#endif				
				else {
#ifndef NDEBUG
					printf("found alltoall of size %i!\n", *si);
#endif
					Collective* nc = new Collective;
					nc->colltype = ALLTOALL;
					nc->tuples = new std::vector<tuple*>;
					for (int i=0; i<num_procs; i++) {
						for (int j=0; j<num_procs; j++) {
							nc->tuples->push_back(alltoall[i][j]);
						}
						RemoveAllTuplesInAfromB(&alltoall[i], tuples);
					}
					colls->push_back(nc);
				}
			}
		}
	}

	{ /******* collective detection - SCATTER (and BROADCAST ) - needs relation $(r, o_r, j, o_j, s)\ 0 <= j < P, j != r$   ******/

		// we need to determine all possible sizes to test for first -- for
		// alltoall, simply use the src=0, dst=0 tuple!
		std::vector<int> candidate_sizes, candidate_roots;
		for(std::vector<tuple*>::iterator ri = tuples->begin(); ri!=tuples->end(); ++ri) {
			tuple *t = *ri;
			candidate_sizes.push_back(t->size);
			candidate_roots.push_back(t->sowner);
		}
		std::sort(candidate_roots.begin(), candidate_roots.end());
		candidate_roots.erase(std::unique(candidate_roots.begin(), candidate_roots.end()), candidate_roots.end());
		std::sort(candidate_sizes.begin(), candidate_sizes.end());
		candidate_sizes.erase(std::unique(candidate_sizes.begin(), candidate_sizes.end()), candidate_sizes.end());

		// here, one should sort the tuples set into a set of sets where
		// each set i only includes the tuples that have rank i as source --
		// one could then get rid of the N^2 runtime of the loop below :)

		for(std::vector<int>::iterator ti = candidate_roots.begin(); ti!=candidate_roots.end(); ++ti) {
			int scatterrootfound=0, gatherrootfound=0; // we need to be able to skip over root if there are no self-sends!
			for(std::vector<int>::iterator si = candidate_sizes.begin(); si!=candidate_sizes.end(); ++si) {
				std::vector<tuple*> scatter, gather;
				int size=0, scatteri=0, gatheri=0;
				int scatterfound = 0;
				int gatherfound = 0;
				scatteri=0; gatheri=0;
				for(std::vector<tuple*>::iterator ri = tuples->begin(); ri!=tuples->end(); ++ri) {
					tuple *t = *ri;

					if(t->owner > *ti && scatterrootfound == 0) {
						scatterrootfound++;
						scatter.push_back(NULL);
						scatterfound++;
						scatteri++;
					}
					//printf("root %i sowner %i scatteri %i, owner %i\n", *ti, t->sowner, scatteri, t->owner);
					if(*ti == t->sowner && scatteri == t->owner && t->size == *si && t->size == t->ssize) {
						scatter.push_back(t);
						scatteri++; scatterfound++;
						if(t->sowner == *ti) scatterrootfound = 1;
					}
					if(t->owner > *ti && gatherrootfound == 0) {
						gatherrootfound++;
						gather.push_back(NULL);
						gatherfound++; gatheri++;
					}
					if(*ti == t->owner && gatheri == t->sowner && t->size == *si && t->size == t->ssize) {
						gather.push_back(t);
						gatheri++; gatherfound++;
						if(t->owner == *ti) gatherrootfound = 1;
					}

				}
#ifndef NDEBUG
				printf("p: %i, size: %i, root, %i, scatterfound %i, gatherfound: %i\n", num_procs, *si, *ti, scatterfound, gatherfound);
#endif
				if(scatterfound >= num_procs) {
					// we found a scatter -- this could, however, also be a
					// broadcast if all contributed buffers from the root are identical!
					int bcast=1;
					uint64_t buf=-1;
					for(int i=0; i<num_procs; ++i) {
						if(scatter[i] != NULL) {
							if(buf == -1) buf = scatter[i]->sstart; // first valid buffer!
							if(buf != scatter[i]->sstart) bcast = 0;
						}
					}
					if(bcast == 1) {
#ifndef NDEBUG
						printf("found bcast from %i of size %i!\n", *ti, *si);
#endif
						Collective* nc = new Collective;
						nc->colltype = BCAST;
						nc->tuples = new std::vector<tuple*>;
						int root = -1;
						uint64_t root_buf = 0; // we need this because of the missing "self send"
						int size = -1;          // tuple - we just infer the information from the other tuples
						int nulltuple_seen = 0;
						for (int i=0; i<scatter.size(); i++) {
							if (scatter[i] != NULL) {
								root = scatter[i]->sowner;
								root_buf = scatter[i]->sstart;
								size = scatter[i]->size;
								nc->tuples->push_back(scatter[i]);
								//tuple* t = scatter.at(i);
								//printf("id: %llu, owner: rank %i, range: (%p,%ib,%i), from: %i, range: (%p,%ib,%i)\n", t->id, t->owner, t->start, t->size, (int)t->mtype, t->sowner, t->sstart, t->ssize, (int)t->smtype);
							}
							else {
								nulltuple_seen = 1;
							}
						}
						if (nulltuple_seen != 0) {
							// add the self send tuple: this is necessary so we generate a call to the
							// MPI funtion even on process root where the bcast does not really move data
							// since tuples only capture data movement no such tuple exists yet.
							assert(size > 0);
							tuple* nt = new tuple;
							nt->id = -1; // shouldn't matter
							nt->owner = root;
							nt->sowner = root;
							nt->size = size;
							nt->ssize = size;
							nt->start = root_buf;
							nt->sstart = root_buf;
							nc->tuples->push_back(nt);
						}
						colls->push_back(nc);
						RemoveAllTuplesInAfromB(&scatter, tuples);
					}
					else {
#ifndef NDEBUG
						printf("found scatter from %i of size %i!\n", *ti, *si);
#endif
						Collective* nc = new Collective;
						nc->colltype = SCATTER;
						nc->tuples = new std::vector<tuple*>;
						int size = -1;
						int root;
						int root_sendbuf; // recvbuf is MPI_IN_PLACE if no self send was detected
						int nulltuple_seen = 0;
						for (int i=0; i<scatter.size(); i++) {
							if (scatter[i] != NULL) {
								size = scatter[i]->size;
								root = scatter[i]->sowner;
								root_sendbuf = scatter[i]->sstart - scatter[i]->size * scatter[i]->owner;
								nc->tuples->push_back(scatter[i]);
								//tuple* t = scatter.at(i);
								//printf("id: %llu, owner: rank %i, range: (%p,%ib,%i), from: %i, range: (%p,%ib,%i)\n", t->id, t->owner, t->start, t->size, (int)t->mtype, t->sowner, t->sstart, t->ssize, (int)t->smtype);
							}
							else {
								nulltuple_seen = 1;
							}
						}
						if (nulltuple_seen != 0) {
							// add the self send tuple: this is necessary so we generate a call to the
							// MPI function even on process root where the scatter does not necessarily move data (MPI_IN_PLACE)
							// since tuples only capture data movement no such tuple exists yet.
							assert(size > 0);
							tuple* nt = new tuple;
							nt->id = -1; // shouldn't matter
							nt->owner = root;
							nt->sowner = root;
							nt->size = size;
							nt->ssize = size;
							nt->start = root_sendbuf;
							nt->sstart = (uint64_t) MPI_IN_PLACE;
							nc->tuples->push_back(nt);
						}
						colls->push_back(nc);
						RemoveAllTuplesInAfromB(&scatter, tuples);
					}
				}
				if(gatherfound == num_procs) {
#ifndef NDEBUG
					printf("found gather to %i of size %i!\n", *ti, *si);
#endif
					Collective* nc = new Collective;
					nc->colltype = GATHER;
					nc->tuples = new std::vector<tuple*>;
					int nulltuple_seen = 0;
					int root = -1;
					uint64_t root_buf = 0; // we need this because of the missing "self send"
					int size = -1;          // tuple - we just infer the information from the other tuples
											// note that if gather has no self send it had to be an MPI_IN_PLACE gather!

					for (int i=0; i<gather.size(); i++) {
						if (gather[i] != NULL) {
							root = gather[i]->owner;
							root_buf = gather[i]->start - gather[i]->size * gather[i]->sowner;
							size = gather[i]->size;
							nc->tuples->push_back(gather[i]);
							tuple* t = gather.at(i);
							//printf("id: %llu, owner: rank %i, range: (%p,%ib,%i), from: %i, range: (%p,%ib,%i)\n", t->id, t->owner, t->start, t->size, (int)t->mtype, t->sowner, t->sstart, t->ssize, (int)t->smtype);
						}
						else {
							nulltuple_seen = 1;
						}
					}
					if (nulltuple_seen != 0) {
						// add the self send tuple: this is necessary so we generate a call to the
						// MPI function even on process root where the gather does not really move data in the MPI_IN_PLACE case
						// since tuples only capture data movement no such tuple exists yet.
						assert(size > 0);
						tuple* nt = new tuple;
						nt->id = -1; // shouldn't matter
						nt->owner = root;
						nt->sowner = root;
						nt->size = size;
						nt->ssize = size;
						nt->start = root_buf;
						nt->sstart = (uint64_t) MPI_IN_PLACE;
						nc->tuples->push_back(nt);
					}
					colls->push_back(nc);
					RemoveAllTuplesInAfromB(&gather, tuples);
				}
			}
		}
	}

	return tuples;

}

